from sklearn import datasets
from sklearn.model_selection import GridSearchCV as SklearnGridSearchCV
from sklearn.model_selection import RandomizedSearchCV \
    as SklearnRandomizedSearchCV
from sklearn.svm import SVC

from skspark.model_selection import GridSearchCV as SparkGridSearchCV
from skspark.model_selection import RandomizedSearchCV \
    as SparkRandomizedSearchCV
from tests.pyspark_test import PySparkTest


class TestDirectComparison(PySparkTest):

    def test_example_grid_search(self):
        # The classic example from the sklearn documentation
        iris = datasets.load_iris()
        parameters = {'kernel': ('linear', 'rbf'), 'C': [1, 10]}
        svr = SVC()
        clf = SklearnGridSearchCV(svr, parameters)
        clf.fit(iris.data, iris.target)

        clf2 = SparkGridSearchCV(self.spark, svr, parameters)
        clf2.fit(iris.data, iris.target)

        b1 = clf.estimator
        b2 = clf2.estimator
        self.assertEqual(b1.get_params(), b2.get_params())

    def test_example_randomized_search(self):
        # The classic example from the sklearn documentation
        iris = datasets.load_iris()
        parameters = {'kernel': ('linear', 'rbf'), 'C': range(1, 10)}
        svr = SVC()
        clf = SklearnRandomizedSearchCV(svr, parameters, random_state=4)
        clf.fit(iris.data, iris.target)

        clf2 = SparkRandomizedSearchCV(self.spark, svr, parameters,
                                       random_state=4)
        clf2.fit(iris.data, iris.target)

        b1 = clf.estimator
        b2 = clf2.estimator
        self.assertEqual(b1.get_params(), b2.get_params())



"""
Python 3 throws ResourceWarning s for unclosed sockets.
There is one test inside sklearn which is supposed to not return any warnings,
but does throw an unrelated resource warning.

This file has this test whilst swallowing the resource warning.
"""
import warnings
import numpy as np

from sklearn.svm import LinearSVC
from sklearn.utils.testing import assert_warns_message, clean_warning_registry

from tests.pyspark_test import PySparkTest
from skspark.model_selection import GridSearchCV, RandomizedSearchCV


def assert_no_warnings(func, *args, **kw):
    # very important to avoid uncontrolled state propagation
    clean_warning_registry()
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')

        result = func(*args, **kw)
        if hasattr(np, 'VisibleDeprecationWarning'):
            # Filter out numpy-specific warnings in numpy >= 1.9
            w = [e for e in w
                 if e.category is not np.VisibleDeprecationWarning]

        # skip ResourceWarning socket warnings
        w = [e for e in w if e.category is not ResourceWarning]

        if len(w) > 0:
            raise AssertionError("Got warnings when calling %s: [%s]"
                                 % (func.__name__,
                                    ', '.join(str(warning) for warning in w)))
    return result


class ResourceWarningTests(PySparkTest):
    """This class contains some tests which break due to returning
    ResourceWarnings."""
    def test_return_train_score_warn(self):
        # Test that warnings are raised. Will be removed in 0.21

        X = np.arange(100).reshape(10, 10)
        y = np.array([0] * 5 + [1] * 5)
        grid = {'C': [1, 2]}

        estimators = [
            GridSearchCV(self.spark, LinearSVC(random_state=0), grid),
            RandomizedSearchCV(
                self.spark, LinearSVC(random_state=0), grid, n_iter=2)
        ]

        result = {}
        for estimator in estimators:
            for val in [True, False, 'warn']:
                estimator.set_params(return_train_score=val)
                result[val] = assert_no_warnings(estimator.fit, X,
                                                 y).cv_results_

        train_keys = ['split0_train_score', 'split1_train_score',
                      'split2_train_score', 'mean_train_score',
                      'std_train_score']
        for key in train_keys:
            msg = (
                'You are accessing a training score ({!r}), '
                'which will not be available by default '
                'any more in 0.21. If you need training scores, '
                'please set return_train_score=True').format(key)
            train_score = assert_warns_message(FutureWarning, msg,
                                               result['warn'].get, key)
            assert np.allclose(train_score, result[True][key])
            assert key not in result[False]

        for key in result['warn']:
            if key not in train_keys:
                assert_no_warnings(result['warn'].get, key)



import unittest
import logging
from pyspark.sql import SparkSession


class PySparkTest(unittest.TestCase):
    """
    Based on this blog post:
    https://blog.cambridgespark.com/unit-testing-with-pyspark-fb31671b1ad8
    """
    @classmethod
    def suppress_py4j_logging(cls):
        logger = logging.getLogger("py4j")
        logger.setLevel(logging.WARN)

    @classmethod
    def create_testing_pyspark_session(cls):
        return SparkSession.builder\
                .master("local[*]")\
                .appName("scikit-spark-tests")\
                .getOrCreate()

    @classmethod
    def setUpClass(cls):
        cls.suppress_py4j_logging()
        cls.spark = cls.create_testing_pyspark_session()

    @classmethod
    def tearDownClass(cls):
        cls.spark.stop()



from types import GeneratorType

import numpy as np
from numpy.testing import assert_array_almost_equal
from sklearn.datasets import make_classification
from sklearn.exceptions import FitFailedWarning
from sklearn.model_selection import KFold
from sklearn.model_selection.tests.common import OneTimeSplitter
from sklearn.model_selection.tests.test_search import MockClassifier, \
    FailingClassifier

from sklearn.svm import LinearSVC, SVC
from sklearn.utils.mocking import CheckingClassifier
from sklearn.utils.testing import assert_no_warnings, assert_warns_message, \
    assert_true, assert_raise_message, assert_raises, assert_warns

from pyspark_test import PySparkTest
from skspark.model_selection import GridSearchCV, RandomizedSearchCV


X = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
y = np.array([1, 1, 2, 2])


def check_hyperparameter_searcher_with_fit_params(spark, klass, **klass_kwargs):
    X = np.arange(100).reshape(10, 10)
    y = np.array([0] * 5 + [1] * 5)
    clf = CheckingClassifier(expected_fit_params=['spam', 'eggs'])
    searcher = klass(spark, clf, {'foo_param': [1, 2, 3]}, cv=2, **klass_kwargs)

    # The CheckingClassifier generates an assertion error if
    # a parameter is missing or has length != len(X).
    assert_raise_message(AssertionError,
                         "Expected fit parameter(s) ['eggs'] not seen.",
                         searcher.fit, X, y, spam=np.ones(10))
    assert_raise_message(AssertionError,
                         "Fit parameter spam has length 1; expected 4.",
                         searcher.fit, X, y, spam=np.ones(1),
                         eggs=np.zeros(10))
    searcher.fit(X, y, spam=np.ones(10), eggs=np.zeros(10))


class TestSpecificRegressionTests(PySparkTest):
    """This class contains some select tests taken from scikit-learn."""
    # def test_return_train_score_warn(self):
    #     # Test that warnings are raised. Will be removed in 0.21
    #
    #     X = np.arange(100).reshape(10, 10)
    #     y = np.array([0] * 5 + [1] * 5)
    #     grid = {'C': [1, 2]}
    #
    #     estimators = [
    #         GridSearchCV(self.spark, LinearSVC(random_state=0), grid),
    #         RandomizedSearchCV(
    #             self.spark, LinearSVC(random_state=0), grid, n_iter=2)
    #     ]
    #
    #     result = {}
    #     for estimator in estimators:
    #         for val in [True, False, 'warn']:
    #             estimator.set_params(return_train_score=val)
    #             result[val] = assert_no_warnings(estimator.fit, X,
    #                                              y).cv_results_
    #
    #     train_keys = ['split0_train_score', 'split1_train_score',
    #                   'split2_train_score', 'mean_train_score',
    #                   'std_train_score']
    #     for key in train_keys:
    #         msg = (
    #             'You are accessing a training score ({!r}), '
    #             'which will not be available by default '
    #             'any more in 0.21. If you need training scores, '
    #             'please set return_train_score=True').format(key)
    #         train_score = assert_warns_message(FutureWarning, msg,
    #                                            result['warn'].get, key)
    #         assert np.allclose(train_score, result[True][key])
    #         assert key not in result[False]
    #
    #     for key in result['warn']:
    #         if key not in train_keys:
    #             assert_no_warnings(result['warn'].get, key)

    def test_trivial_cv_results_attr(self):
        # Test search over a "grid" with only one point.
        # Non-regression test: grid_scores_ wouldn't be set by GridSearchCV.
        clf = MockClassifier()
        grid_search = GridSearchCV(self.spark, clf, {'foo_param': [1]})
        grid_search.fit(X, y)
        assert_true(hasattr(grid_search, "cv_results_"))

        # random_search = RandomizedSearchCV(clf, {'foo_param': [0]}, n_iter=1)
        # random_search.fit(X, y)
        # assert_true(hasattr(grid_search, "cv_results_"))

    def test_random_search_with_fit_params(self):
        check_hyperparameter_searcher_with_fit_params(
            self.spark, RandomizedSearchCV, n_iter=1)

    def test_grid_search_precomputed_kernel_error_nonsquare(self):
        # Test that grid search returns an error with a non-square precomputed
        # training kernel matrix
        K_train = np.zeros((10, 20))
        y_train = np.ones((10,))
        clf = SVC(kernel='precomputed')
        cv = GridSearchCV(self.spark, clf, {'C': [0.1, 1.0]})
        assert_raises(ValueError, cv.fit, K_train, y_train)

    def test_grid_search_cv_splits_consistency(self):
        # Check if a one time iterable is accepted as a cv parameter.
        n_samples = 100
        n_splits = 5
        X, y = make_classification(n_samples=n_samples, random_state=0)

        gs = GridSearchCV(self.spark, LinearSVC(random_state=0),
                          param_grid={'C': [0.1, 0.2, 0.3]},
                          cv=OneTimeSplitter(n_splits=n_splits,
                                             n_samples=n_samples))
        gs.fit(X, y)

        gs2 = GridSearchCV(self.spark, LinearSVC(random_state=0),
                           param_grid={'C': [0.1, 0.2, 0.3]},
                           cv=KFold(n_splits=n_splits))
        gs2.fit(X, y)

        # Give generator as a cv parameter
        assert_true(isinstance(KFold(n_splits=n_splits,
                                     shuffle=True, random_state=0).split(X, y),
                               GeneratorType))
        gs3 = GridSearchCV(self.spark, LinearSVC(random_state=0),
                           param_grid={'C': [0.1, 0.2, 0.3]},
                           cv=KFold(n_splits=n_splits, shuffle=True,
                                    random_state=0).split(X, y))
        gs3.fit(X, y)

        gs4 = GridSearchCV(self.spark, LinearSVC(random_state=0),
                           param_grid={'C': [0.1, 0.2, 0.3]},
                           cv=KFold(n_splits=n_splits, shuffle=True,
                                    random_state=0))
        gs4.fit(X, y)

        def _pop_time_keys(cv_results):
            for key in ('mean_fit_time', 'std_fit_time',
                        'mean_score_time', 'std_score_time'):
                cv_results.pop(key)
            return cv_results

        # Check if generators are supported as cv and
        # that the splits are consistent
        np.testing.assert_equal(_pop_time_keys(gs3.cv_results_),
                                _pop_time_keys(gs4.cv_results_))

        # OneTimeSplitter is a non-re-entrant cv where split can be called only
        # once if ``cv.split`` is called once per param setting in GridSearchCV.fit
        # the 2nd and 3rd parameter will not be evaluated as no train/test indices
        # will be generated for the 2nd and subsequent cv.split calls.
        # This is a check to make sure cv.split is not called once per param
        # setting.
        np.testing.assert_equal({k: v for k, v in gs.cv_results_.items()
                                 if not k.endswith('_time')},
                                {k: v for k, v in gs2.cv_results_.items()
                                 if not k.endswith('_time')})

        # Check consistency of folds across the parameters
        gs = GridSearchCV(self.spark, LinearSVC(random_state=0),
                          param_grid={'C': [0.1, 0.1, 0.2, 0.2]},
                          cv=KFold(n_splits=n_splits, shuffle=True))
        gs.fit(X, y)

        # As the first two param settings (C=0.1) and the next two param
        # settings (C=0.2) are same, the test and train scores must also be
        # same as long as the same train/test indices are generated for all
        # the cv splits, for both param setting
        for score_type in ('train', 'test'):
            per_param_scores = {}
            for param_i in range(4):
                per_param_scores[param_i] = list(
                    gs.cv_results_['split%d_%s_score' % (s, score_type)][
                        param_i]
                    for s in range(5))

            assert_array_almost_equal(per_param_scores[0],
                                      per_param_scores[1])
            assert_array_almost_equal(per_param_scores[2],
                                      per_param_scores[3])

    def test_grid_search_failing_classifier(self):
        # GridSearchCV with on_error != 'raise'
        # Ensures that a warning is raised and score reset where appropriate.

        X, y = make_classification(n_samples=20, n_features=10, random_state=0)

        clf = FailingClassifier()

        # refit=False because we only want to check that errors caused by fits
        # to individual folds will be caught and warnings raised instead. If
        # refit was done, then an exception would be raised on refit and not
        # caught by grid_search (expected behavior), and this would cause an
        # error in this test.
        gs = GridSearchCV(self.spark, clf, [{'parameter': [0, 1, 2]}],
                          scoring='accuracy', refit=False, error_score=0.0)

        assert_warns(FitFailedWarning, gs.fit, X, y)

        # Ensure that grid scores were set to zero as required for those fits
        # that are expected to fail.
        assert all(np.all(this_point.cv_validation_scores == 0.0)
                   for this_point in gs.grid_scores_
                   if this_point.parameters['parameter'] ==
                   FailingClassifier.FAILING_PARAMETER)

        gs = GridSearchCV(self.spark, clf, [{'parameter': [0, 1, 2]}],
                          scoring='accuracy', refit=False,
                          error_score=float('nan'))
        assert_warns(FitFailedWarning, gs.fit, X, y)
        assert all(np.all(np.isnan(this_point.cv_validation_scores))
                   for this_point in gs.grid_scores_
                   if this_point.parameters['parameter'] ==
                   FailingClassifier.FAILING_PARAMETER)



import sklearn.model_selection
from skspark.model_selection import RandomizedSearchCV as \
    SparkRandomizedSearchCV
from skspark.model_selection import GridSearchCV as SparkGridSearchCV

from sklearn.model_selection import RandomizedSearchCV as \
    SklearnRandomizedSearchCV
from sklearn.model_selection import GridSearchCV as SklearnGridSearchCV

from tests.pyspark_test import PySparkTest


# Overwrite the sklearn GridSearch in this suite so that we can run the same
# tests with the same parameters.
class AllTests(PySparkTest):
    # After testing, make sure to revert sklearn to normal
    # (see _add_to_module())
    @classmethod
    def tearDownClass(cls):
        super(AllTests, cls).tearDownClass()
        # Restore sklearn module to the original state after done testing
        # this fixture.
        sklearn.model_selection.GridSearchCV = \
            sklearn.model_selection.GridSearchCV_original
        del sklearn.model_selection.GridSearchCV_original

        sklearn.model_selection.RandomizedSearchCV = \
            sklearn.model_selection.RandomizedSearchCV_original
        del sklearn.model_selection.RandomizedSearchCV_original


class GridSearchCV(SparkGridSearchCV):
    """
    Wrapper to make the scikit-spark version of GridSearchCV seem like
    the scikit-learn version. This allows it to be used in the
    existing scikit-learn tests.
    """
    def __init__(self, estimator, param_grid, scoring=None, fit_params=None,
                 n_jobs=1, iid='warn', refit=True, cv=None, verbose=0,
                 pre_dispatch='2*n_jobs', error_score='raise',
                 return_train_score="warn"):
        super(GridSearchCV, self).__init__(
            AllTests.spark, estimator, param_grid, scoring, fit_params, n_jobs,
            iid, refit, cv, verbose, pre_dispatch, error_score,
            return_train_score)


class RandomizedSearchCV(SparkRandomizedSearchCV):
    """
    Wrapper to make the scikit-spark version of RandomizedSearchCV seem like
    the scikit-learn version. This allows it to be used in the
    existing scikit-learn tests.
    """
    def __init__(self, estimator, param_distributions, n_iter=10,
                 scoring=None, fit_params=None, n_jobs=1, iid='warn',
                 refit=True, cv=None, verbose=0, pre_dispatch='2*n_jobs',
                 random_state=None, error_score='raise',
                 return_train_score="warn"):
        super(RandomizedSearchCV, self).__init__(
            AllTests.spark, estimator, param_distributions, n_iter, scoring,
            fit_params, n_jobs, iid, refit, cv, verbose, pre_dispatch,
            random_state, error_score, return_train_score)


def _create_method(method):
    def do_test_expected(*kwargs):
        method()

    return do_test_expected


def _add_to_module():
    sklearn.model_selection.GridSearchCV = GridSearchCV
    sklearn.model_selection.GridSearchCV_original = SklearnGridSearchCV

    sklearn.model_selection.RandomizedSearchCV = RandomizedSearchCV
    sklearn.model_selection.RandomizedSearchCV_original = \
        SklearnRandomizedSearchCV

    from sklearn.model_selection.tests import test_search
    all_tests = [
        test
        for (method_name, test) in test_search.__dict__.items()
        if method_name.startswith("test_")
    ]

    # These tests have been edited and moved into this repo e.g. in
    # resource_warning_tests.py
    refactored_tests = ["test_return_train_score_warn"]

    for test in all_tests:
        if test.__name__ not in refactored_tests:
            test_to_add = _create_method(test)
            setattr(AllTests, test.__name__, test_to_add)


_add_to_module()



import warnings
from collections import defaultdict
from functools import partial

import six
import numpy as np
from itertools import product
from sklearn.base import is_classifier, clone
from sklearn.externals.joblib import Parallel, delayed
from sklearn.metrics.scorer import _check_multimetric_scoring
from sklearn.model_selection import check_cv
from sklearn.model_selection._validation import _fit_and_score, \
    _aggregate_score_dicts
from sklearn.utils import indexable
from sklearn.model_selection._search import BaseSearchCV
from sklearn.utils.deprecation import DeprecationDict
from sklearn.utils.fixes import MaskedArray
from scipy.stats import rankdata


class SparkBaseSearchCV(BaseSearchCV):
    def __init__(self, spark, estimator, scoring=None, fit_params=None,
                 n_jobs=1, iid=True, refit=True, cv=None, verbose=0,
                 pre_dispatch='2*n_jobs', error_score='raise',
                 return_train_score=True):

        super(SparkBaseSearchCV, self).__init__(
            estimator, scoring, fit_params, n_jobs, iid, refit, cv, verbose,
            pre_dispatch, error_score, return_train_score)

        self.spark = spark

        # TODO: I'm not doing whatever pre_dispatch is supposed to do

    def fit(self, X, y=None, groups=None, **fit_params):
        """Run fit with all sets of parameters.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples] or [n_samples, n_output], optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of the estimator
        """
        if self.fit_params is not None:
            warnings.warn('"fit_params" as a constructor argument was '
                          'deprecated in version 0.19 and will be removed '
                          'in version 0.21. Pass fit parameters to the '
                          '"fit" method instead.', DeprecationWarning)
            if fit_params:
                warnings.warn('Ignoring fit_params passed as a constructor '
                              'argument in favor of keyword arguments to '
                              'the "fit" method.', RuntimeWarning)
            else:
                fit_params = self.fit_params
        estimator = self.estimator
        cv = check_cv(self.cv, y, classifier=is_classifier(estimator))

        scorers, self.multimetric_ = _check_multimetric_scoring(
            self.estimator, scoring=self.scoring)

        if self.multimetric_:
            if self.refit is not False and (
                    not isinstance(self.refit, six.string_types) or
                    # This will work for both dict / list (tuple)
                    self.refit not in scorers):
                raise ValueError("For multi-metric scoring, the parameter "
                                 "refit must be set to a scorer key "
                                 "to refit an estimator with the best "
                                 "parameter setting on the whole data and "
                                 "make the best_* attributes "
                                 "available for that metric. If this is not "
                                 "needed, refit should be set to False "
                                 "explicitly. %r was passed." % self.refit)
            else:
                refit_metric = self.refit
        else:
            refit_metric = 'score'

        X, y, groups = indexable(X, y, groups)
        n_splits = cv.get_n_splits(X, y, groups)
        # Regenerate parameter iterable for each fit
        candidate_params = list(self._get_param_iterator())
        n_candidates = len(candidate_params)
        if self.verbose > 0:
            print("Fitting {0} folds for each of {1} candidates, totalling"
                  " {2} fits".format(n_splits, n_candidates,
                                     n_candidates * n_splits))

        base_estimator = clone(self.estimator)

        method = "spark"
        if method == "sklearn":
            out = self._run_sklearn_fit(base_estimator, X, y, scorers,
                                        fit_params, candidate_params, cv,
                                        groups)
        elif method == "spark":
            out = self._run_skspark_fit(base_estimator, X, y, scorers,
                                        fit_params, candidate_params, cv,
                                        groups)

        # if one choose to see train score, "out" will contain train score info
        if self.return_train_score:
            (train_score_dicts, test_score_dicts, test_sample_counts, fit_time,
             score_time) = zip(*out)
        else:
            (test_score_dicts, test_sample_counts, fit_time,
             score_time) = zip(*out)

        # test_score_dicts and train_score dicts are lists of dictionaries and
        # we make them into dict of lists
        test_scores = _aggregate_score_dicts(test_score_dicts)
        if self.return_train_score:
            train_scores = _aggregate_score_dicts(train_score_dicts)

        # TODO: replace by a dict in 0.21
        results = (DeprecationDict() if self.return_train_score == 'warn'
                   else {})

        def _store(key_name, array, weights=None, splits=False, rank=False):
            """A small helper to store the scores/times to the cv_results_"""
            # When iterated first by splits, then by parameters
            # We want `array` to have `n_candidates` rows and `n_splits` cols.
            array = np.array(array, dtype=np.float64).reshape(n_candidates,
                                                              n_splits)
            if splits:
                for split_i in range(n_splits):
                    # Uses closure to alter the results
                    results["split%d_%s"
                            % (split_i, key_name)] = array[:, split_i]

            array_means = np.average(array, axis=1, weights=weights)
            results['mean_%s' % key_name] = array_means
            # Weighted std is not directly available in numpy
            array_stds = np.sqrt(np.average((array -
                                             array_means[:, np.newaxis]) ** 2,
                                            axis=1, weights=weights))
            results['std_%s' % key_name] = array_stds

            if rank:
                results["rank_%s" % key_name] = np.asarray(
                    rankdata(-array_means, method='min'), dtype=np.int32)

        _store('fit_time', fit_time)
        _store('score_time', score_time)
        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results = defaultdict(partial(MaskedArray,
                                            np.empty(n_candidates, ),
                                            mask=True,
                                            dtype=object))
        for cand_i, params in enumerate(candidate_params):
            for name, value in params.items():
                # An all masked empty array gets created for the key
                # `"param_%s" % name` at the first occurence of `name`.
                # Setting the value at an index also unmasks that index
                param_results["param_%s" % name][cand_i] = value

        results.update(param_results)
        # Store a list of param dicts at the key 'params'
        results['params'] = candidate_params

        # NOTE test_sample counts (weights) remain the same for all candidates
        test_sample_counts = np.array(test_sample_counts[:n_splits],
                                      dtype=np.int)
        for scorer_name in scorers.keys():
            # Computed the (weighted) mean and std for test scores alone
            _store('test_%s' % scorer_name, test_scores[scorer_name],
                   splits=True, rank=True,
                   weights=test_sample_counts if self.iid else None)
            if self.return_train_score:
                prev_keys = set(results.keys())
                _store('train_%s' % scorer_name, train_scores[scorer_name],
                       splits=True)

                if self.return_train_score == 'warn':
                    for key in set(results.keys()) - prev_keys:
                        message = (
                            'You are accessing a training score ({!r}), '
                            'which will not be available by default '
                            'any more in 0.21. If you need training scores, '
                            'please set return_train_score=True').format(key)
                        # warn on key access
                        results.add_warning(key, message, FutureWarning)

        # For multi-metric evaluation, store the best_index_, best_params_ and
        # best_score_ iff refit is one of the scorer names
        # In single metric evaluation, refit_metric is "score"
        if self.refit or not self.multimetric_:
            self.best_index_ = results["rank_test_%s" % refit_metric].argmin()
            self.best_params_ = candidate_params[self.best_index_]
            self.best_score_ = results["mean_test_%s" % refit_metric][
                self.best_index_]

        if self.refit:
            self.best_estimator_ = clone(base_estimator).set_params(
                **self.best_params_)
            if y is not None:
                self.best_estimator_.fit(X, y, **fit_params)
            else:
                self.best_estimator_.fit(X, **fit_params)

        # Store the only scorer not as a dict for single metric evaluation
        self.scorer_ = scorers if self.multimetric_ else scorers['score']

        self.cv_results_ = results
        self.n_splits_ = n_splits

        return self

    def _run_sklearn_fit(self, base_estimator, X, y, scorers, fit_params,
                         candidate_params, cv, groups):
        """Run fitting locally"""
        pre_dispatch = self.pre_dispatch

        out = Parallel(
            n_jobs=self.n_jobs, verbose=self.verbose,
            pre_dispatch=pre_dispatch
        )(delayed(_fit_and_score)(clone(base_estimator), X, y, scorers, train,
                                  test, self.verbose, parameters,
                                  fit_params=fit_params,
                                  return_train_score=self.return_train_score,
                                  return_n_test_samples=True,
                                  return_times=True, return_parameters=False,
                                  error_score=self.error_score)
          for parameters, (train, test) in product(candidate_params,
                                                   cv.split(X, y, groups)))

        return out

    def _run_skspark_fit(self, base_estimator, X, y, scorers, fit_params,
                         candidate_params, cv, groups):
        param_grid = [(parameters, train, test) for parameters, (train, test)
                      in product(candidate_params, cv.split(X, y, groups))]

        # Because the original python code expects a certain order for the
        # elements, we need to respect it.
        indexed_param_grid = list(zip(range(len(param_grid)), param_grid))
        par_param_grid = self.spark.sparkContext.parallelize(
            indexed_param_grid, len(indexed_param_grid))

        X_bc = self.spark.sparkContext.broadcast(X)
        y_bc = self.spark.sparkContext.broadcast(y)

        verbose = self.verbose
        error_score = self.error_score
        return_train_score = self.return_train_score

        def spark_task(tup):
            (index, (parameters, train, test)) = tup
            local_estimator = clone(base_estimator)
            local_X = X_bc.value
            local_y = y_bc.value

            error = None
            with warnings.catch_warnings(record=True) as warns:
                try:
                    res = _fit_and_score(
                        local_estimator, local_X, local_y, scorers, train, test,
                        verbose, parameters, fit_params,
                        return_train_score=return_train_score,
                        return_parameters=True,
                        return_n_test_samples=True, return_times=True,
                        error_score=error_score)
                except Exception as e:
                    res = None
                    error = e

            return index, {"results": res, "errors": error, "warnings": warns}

        indexed_out0 = dict(par_param_grid.map(spark_task).collect())

        # process errors and warnings
        for i in range(len(param_grid)):
            error_to_be_raised = indexed_out0[i]["errors"]
            if error_to_be_raised is not None:
                raise error_to_be_raised

            for warn in indexed_out0[i]["warnings"]:
                warnings.warn(warn.message, warn.category)

        out = [indexed_out0[i]["results"][:-1] for i in range(len(param_grid))]

        X_bc.unpersist()
        y_bc.unpersist()
        return out

    def __getstate__(self):
        """To not try to pickle the non-serializable SparkSession"""
        attributes = dict(self.__dict__)
        del attributes['spark']
        return attributes



from sklearn.model_selection import ParameterGrid, ParameterSampler
from sklearn.model_selection._search import _check_param_grid

from skspark.spark_base_search_cv import SparkBaseSearchCV


class GridSearchCV(SparkBaseSearchCV):
    """Exhaustive search over specified parameter values for an estimator.

    Important members are fit, predict.

    GridSearchCV implements a "fit" and a "score" method.
    It also implements "predict", "predict_proba", "decision_function",
    "transform" and "inverse_transform" if they are implemented in the
    estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated grid-search over a parameter grid.

    Read more in the :ref:`User Guide <grid_search>`.

    Parameters
    ----------
    estimator : estimator object.
        This is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.

    param_grid : dict or list of dictionaries
        Dictionary with parameters names (string) as keys and lists of
        parameter settings to try as values, or a list of such
        dictionaries, in which case the grids spanned by each dictionary
        in the list are explored. This enables searching over any sequence
        of parameter settings.

    scoring : string, callable, list/tuple, dict or None, default: None
        A single string (see :ref:`scoring_parameter`) or a callable
        (see :ref:`scoring`) to evaluate the predictions on the test set.

        For evaluating multiple metrics, either give a list of (unique) strings
        or a dict with names as keys and callables as values.

        NOTE that when using custom scorers, each scorer should return a single
        value. Metric functions returning a list/array of values can be wrapped
        into multiple scorers that return one value each.

        See :ref:`multimetric_grid_search` for an example.

        If None, the estimator's default scorer (if available) is used.

    fit_params : dict, optional
        Parameters to pass to the fit method.

        .. deprecated:: 0.19
           ``fit_params`` as a constructor argument was deprecated in version
           0.19 and will be removed in version 0.21. Pass fit parameters to
           the ``fit`` method instead.

    n_jobs : int, default=1
        Number of jobs to run in parallel.

    pre_dispatch : int, or string, optional
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    iid : boolean, default=True
        If True, the data is assumed to be identically distributed across
        the folds, and the loss minimized is the total loss per sample,
        and not the mean loss across the folds.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
          - None, to use the default 3-fold cross validation,
          - integer, to specify the number of folds in a `(Stratified)KFold`,
          - An object to be used as a cross-validation generator.
          - An iterable yielding train, test splits.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

    refit : boolean, or string, default=True
        Refit an estimator using the best found parameters on the whole
        dataset.

        For multiple metric evaluation, this needs to be a string denoting the
        scorer is used to find the best parameters for refitting the estimator
        at the end.

        The refitted estimator is made available at the ``best_estimator_``
        attribute and permits using ``predict`` directly on this
        ``GridSearchCV`` instance.

        Also for multiple metric evaluation, the attributes ``best_index_``,
        ``best_score_`` and ``best_parameters_`` will only be available if
        ``refit`` is set and all of them will be determined w.r.t this specific
        scorer.

        See ``scoring`` parameter to know more about multiple metric
        evaluation.

    verbose : integer
        Controls the verbosity: the higher, the more messages.

    error_score : 'raise' (default) or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error.

    return_train_score : boolean, optional
        If ``False``, the ``cv_results_`` attribute will not include training
        scores.

        Current default is ``'warn'``, which behaves as ``True`` in addition
        to raising a warning when a training score is looked up.
        That default will be changed to ``False`` in 0.21.
        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.


    Examples
    --------
    >>> from sklearn import svm, datasets
    >>> from skspark.model_selection import GridSearchCV
    >>> from pyspark.sql import SparkSession
    >>>
    >>> iris = datasets.load_iris()
    >>> parameters = {'kernel':('linear', 'rbf'), 'C':[1, 10]}
    >>> svc = svm.SVC()
    >>>
    >>> spark = SparkSession.builder\
            .master("local[*]")\
            .appName("skspark-grid-search-doctests")\
            .getOrCreate()
    >>> clf = GridSearchCV(spark, svc, parameters)
    >>> clf.fit(iris.data, iris.target)
    ...                             # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    GridSearchCV(cv=None, error_score=...,
           estimator=SVC(C=1.0, cache_size=..., class_weight=..., coef0=...,
                         decision_function_shape='ovr', degree=..., gamma=...,
                         kernel='rbf', max_iter=-1, probability=False,
                         random_state=None, shrinking=True, tol=...,
                         verbose=False),
           fit_params=None, iid=..., n_jobs=1,
           param_grid=..., pre_dispatch=..., refit=..., return_train_score=...,
           scoring=..., verbose=...)
    >>> sorted(clf.cv_results_.keys())
    ...                             # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    ['mean_fit_time', 'mean_score_time', 'mean_test_score',...
     'mean_train_score', 'param_C', 'param_kernel', 'params',...
     'rank_test_score', 'split0_test_score',...
     'split0_train_score', 'split1_test_score', 'split1_train_score',...
     'split2_test_score', 'split2_train_score',...
     'std_fit_time', 'std_score_time', 'std_test_score', 'std_train_score'...]

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.

        For instance the below given table

        +------------+-----------+------------+-----------------+---+---------+
        |param_kernel|param_gamma|param_degree|split0_test_score|...|rank_t...|
        +============+===========+============+=================+===+=========+
        |  'poly'    |     --    |      2     |        0.8      |...|    2    |
        +------------+-----------+------------+-----------------+---+---------+
        |  'poly'    |     --    |      3     |        0.7      |...|    4    |
        +------------+-----------+------------+-----------------+---+---------+
        |  'rbf'     |     0.1   |     --     |        0.8      |...|    3    |
        +------------+-----------+------------+-----------------+---+---------+
        |  'rbf'     |     0.2   |     --     |        0.9      |...|    1    |
        +------------+-----------+------------+-----------------+---+---------+

        will be represented by a ``cv_results_`` dict of::

            {
            'param_kernel': masked_array(data = ['poly', 'poly', 'rbf', 'rbf'],
                                         mask = [False False False False]...)
            'param_gamma': masked_array(data = [-- -- 0.1 0.2],
                                        mask = [ True  True False False]...),
            'param_degree': masked_array(data = [2.0 3.0 -- --],
                                         mask = [False False  True  True]...),
            'split0_test_score'  : [0.8, 0.7, 0.8, 0.9],
            'split1_test_score'  : [0.82, 0.5, 0.7, 0.78],
            'mean_test_score'    : [0.81, 0.60, 0.75, 0.82],
            'std_test_score'     : [0.02, 0.01, 0.03, 0.03],
            'rank_test_score'    : [2, 4, 3, 1],
            'split0_train_score' : [0.8, 0.9, 0.7],
            'split1_train_score' : [0.82, 0.5, 0.7],
            'mean_train_score'   : [0.81, 0.7, 0.7],
            'std_train_score'    : [0.03, 0.03, 0.04],
            'mean_fit_time'      : [0.73, 0.63, 0.43, 0.49],
            'std_fit_time'       : [0.01, 0.02, 0.01, 0.01],
            'mean_score_time'    : [0.007, 0.06, 0.04, 0.04],
            'std_score_time'     : [0.001, 0.002, 0.003, 0.005],
            'params'             : [{'kernel': 'poly', 'degree': 2}, ...],
            }

        NOTE

        The key ``'params'`` is used to store a list of parameter
        settings dicts for all the parameter candidates.

        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.

        For multi-metric evaluation, the scores for all the scorers are
        available in the ``cv_results_`` dict at the keys ending with that
        scorer's name (``'_<scorer_name>'``) instead of ``'_score'`` shown
        above. ('split0_test_precision', 'mean_train_precision' etc.)

    best_estimator_ : estimator or dict
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if ``refit=False``.

        See ``refit`` parameter for more information on allowed values.

    best_score_ : float
        Mean cross-validated score of the best_estimator

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    scorer_ : function or a dict
        Scorer function used on the held out data to choose the best
        parameters for the model.

        For multi-metric evaluation, this attribute holds the validated
        ``scoring`` dict which maps the scorer key to the scorer callable.

    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    Notes
    ------
    The parameters selected are those that maximize the score of the left out
    data, unless an explicit score is passed in which case it is used instead.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    point in the grid (and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is `2 *
    n_jobs`.

    See Also
    ---------
    :class:`ParameterGrid`:
        generates all the combinations of a hyperparameter grid.

    :func:`sklearn.model_selection.train_test_split`:
        utility function to split the data into a development set usable
        for fitting a GridSearchCV instance and an evaluation set for
        its final evaluation.

    :func:`sklearn.metrics.make_scorer`:
        Make a scorer from a performance metric or loss function.

    """

    def __init__(self, spark, estimator, param_grid, scoring=None,
                 fit_params=None, n_jobs=1, iid=True, refit=True, cv=None,
                 verbose=0, pre_dispatch='2*n_jobs', error_score='raise',
                 return_train_score="warn"):
        super(GridSearchCV, self).__init__(
            spark=spark, estimator=estimator, scoring=scoring,
            fit_params=fit_params, n_jobs=n_jobs, iid=iid, refit=refit, cv=cv,
            verbose=verbose, pre_dispatch=pre_dispatch, error_score=error_score,
            return_train_score=return_train_score)
        self.param_grid = param_grid
        _check_param_grid(param_grid)

    def _get_param_iterator(self):
        """Return ParameterGrid instance for the given param_grid"""
        return ParameterGrid(self.param_grid)


class RandomizedSearchCV(SparkBaseSearchCV):
    """Randomized search on hyper parameters.

    RandomizedSearchCV implements a "fit" and a "score" method.
    It also implements "predict", "predict_proba", "decision_function",
    "transform" and "inverse_transform" if they are implemented in the
    estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated search over parameter settings.

    In contrast to GridSearchCV, not all parameter values are tried out, but
    rather a fixed number of parameter settings is sampled from the specified
    distributions. The number of parameter settings that are tried is
    given by n_iter.

    If all parameters are presented as a list,
    sampling without replacement is performed. If at least one parameter
    is given as a distribution, sampling with replacement is used.
    It is highly recommended to use continuous distributions for continuous
    parameters.

    Read more in the :ref:`User Guide <randomized_parameter_search>`.

    Parameters
    ----------
    estimator : estimator object.
        A object of that type is instantiated for each grid point.
        This is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.

    param_distributions : dict
        Dictionary with parameters names (string) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        If a list is given, it is sampled uniformly.

    n_iter : int, default=10
        Number of parameter settings that are sampled. n_iter trades
        off runtime vs quality of the solution.

    scoring : string, callable, list/tuple, dict or None, default: None
        A single string (see :ref:`scoring_parameter`) or a callable
        (see :ref:`scoring`) to evaluate the predictions on the test set.

        For evaluating multiple metrics, either give a list of (unique) strings
        or a dict with names as keys and callables as values.

        NOTE that when using custom scorers, each scorer should return a single
        value. Metric functions returning a list/array of values can be wrapped
        into multiple scorers that return one value each.

        See :ref:`multimetric_grid_search` for an example.

        If None, the estimator's default scorer (if available) is used.

    fit_params : dict, optional
        Parameters to pass to the fit method.

        .. deprecated:: 0.19
           ``fit_params`` as a constructor argument was deprecated in version
           0.19 and will be removed in version 0.21. Pass fit parameters to
           the ``fit`` method instead.

    n_jobs : int, default=1
        Number of jobs to run in parallel.

    pre_dispatch : int, or string, optional
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    iid : boolean, default=True
        If True, the data is assumed to be identically distributed across
        the folds, and the loss minimized is the total loss per sample,
        and not the mean loss across the folds.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
          - None, to use the default 3-fold cross validation,
          - integer, to specify the number of folds in a `(Stratified)KFold`,
          - An object to be used as a cross-validation generator.
          - An iterable yielding train, test splits.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

    refit : boolean, or string default=True
        Refit an estimator using the best found parameters on the whole
        dataset.

        For multiple metric evaluation, this needs to be a string denoting the
        scorer that would be used to find the best parameters for refitting
        the estimator at the end.

        The refitted estimator is made available at the ``best_estimator_``
        attribute and permits using ``predict`` directly on this
        ``RandomizedSearchCV`` instance.

        Also for multiple metric evaluation, the attributes ``best_index_``,
        ``best_score_`` and ``best_parameters_`` will only be available if
        ``refit`` is set and all of them will be determined w.r.t this specific
        scorer.

        See ``scoring`` parameter to know more about multiple metric
        evaluation.

    verbose : integer
        Controls the verbosity: the higher, the more messages.

    random_state : int, RandomState instance or None, optional, default=None
        Pseudo random number generator state used for random uniform sampling
        from lists of possible values instead of scipy.stats distributions.
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    error_score : 'raise' (default) or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error.

    return_train_score : boolean, optional
        If ``False``, the ``cv_results_`` attribute will not include training
        scores.

        Current default is ``'warn'``, which behaves as ``True`` in addition
        to raising a warning when a training score is looked up.
        That default will be changed to ``False`` in 0.21.
        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.

        For instance the below given table

        +--------------+-------------+-------------------+---+---------------+
        | param_kernel | param_gamma | split0_test_score |...|rank_test_score|
        +==============+=============+===================+===+===============+
        |    'rbf'     |     0.1     |        0.8        |...|       2       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.2     |        0.9        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.3     |        0.7        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+

        will be represented by a ``cv_results_`` dict of::

            {
            'param_kernel' : masked_array(data = ['rbf', 'rbf', 'rbf'],
                                          mask = False),
            'param_gamma'  : masked_array(data = [0.1 0.2 0.3], mask = False),
            'split0_test_score'  : [0.8, 0.9, 0.7],
            'split1_test_score'  : [0.82, 0.5, 0.7],
            'mean_test_score'    : [0.81, 0.7, 0.7],
            'std_test_score'     : [0.02, 0.2, 0.],
            'rank_test_score'    : [3, 1, 1],
            'split0_train_score' : [0.8, 0.9, 0.7],
            'split1_train_score' : [0.82, 0.5, 0.7],
            'mean_train_score'   : [0.81, 0.7, 0.7],
            'std_train_score'    : [0.03, 0.03, 0.04],
            'mean_fit_time'      : [0.73, 0.63, 0.43, 0.49],
            'std_fit_time'       : [0.01, 0.02, 0.01, 0.01],
            'mean_score_time'    : [0.007, 0.06, 0.04, 0.04],
            'std_score_time'     : [0.001, 0.002, 0.003, 0.005],
            'params'             : [{'kernel' : 'rbf', 'gamma' : 0.1}, ...],
            }

        NOTE

        The key ``'params'`` is used to store a list of parameter
        settings dicts for all the parameter candidates.

        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.

        For multi-metric evaluation, the scores for all the scorers are
        available in the ``cv_results_`` dict at the keys ending with that
        scorer's name (``'_<scorer_name>'``) instead of ``'_score'`` shown
        above. ('split0_test_precision', 'mean_train_precision' etc.)

    best_estimator_ : estimator or dict
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if ``refit=False``.

        For multi-metric evaluation, this attribute is present only if
        ``refit`` is specified.

        See ``refit`` parameter for more information on allowed values.

    best_score_ : float
        Mean cross-validated score of the best_estimator.

        For multi-metric evaluation, this is not available if ``refit`` is
        ``False``. See ``refit`` parameter for more information.

    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.

        For multi-metric evaluation, this is not available if ``refit`` is
        ``False``. See ``refit`` parameter for more information.

    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

        For multi-metric evaluation, this is not available if ``refit`` is
        ``False``. See ``refit`` parameter for more information.

    scorer_ : function or a dict
        Scorer function used on the held out data to choose the best
        parameters for the model.

        For multi-metric evaluation, this attribute holds the validated
        ``scoring`` dict which maps the scorer key to the scorer callable.

    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    Notes
    -----
    The parameters selected are those that maximize the score of the held-out
    data, according to the scoring parameter.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    parameter setting(and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is `2 *
    n_jobs`.

    See Also
    --------
    :class:`GridSearchCV`:
        Does exhaustive search over a grid of parameters.

    :class:`ParameterSampler`:
        A generator over parameter settins, constructed from
        param_distributions.

    """

    def __init__(self, spark, estimator, param_distributions, n_iter=10,
                 scoring=None, fit_params=None, n_jobs=1, iid=True, refit=True,
                 cv=None, verbose=0, pre_dispatch='2*n_jobs', random_state=None,
                 error_score='raise', return_train_score="warn"):

        self.param_distributions = param_distributions
        self.n_iter = n_iter
        self.random_state = random_state

        super(RandomizedSearchCV, self).__init__(
            spark=spark, estimator=estimator, scoring=scoring,
            fit_params=fit_params, n_jobs=n_jobs, iid=iid, refit=refit, cv=cv,
            verbose=verbose, pre_dispatch=pre_dispatch, error_score=error_score,
            return_train_score=return_train_score)

    def _get_param_iterator(self):
        """Return ParameterSampler instance for the given distributions"""
        return ParameterSampler(
            self.param_distributions, self.n_iter,
            random_state=self.random_state)



from setuptools import setup, find_packages

description = "Spark acceleration for Scikit-Learn cross validation techniques"

keywords = [
    "spark",
    "pyspark",
    "scikit-learn",
    "sklearn",
    "machine learning",
    "random search",
    "grid search"
]

install_requires = [
    "numpy>=1.13.0",
    "scikit-learn>=0.19.1",
    "six==1.11.0"
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="scikit-spark",
    version="0.1.0rc4",
    author="Ganesh N. Sivalingam",
    author_email="g.n.sivalingam@gmail.com",
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=keywords,
    package_dir={"": "python"},
    packages=find_packages("python", exclude="tests"),
    url="https://github.com/scikit-spark/scikit-spark",
    install_requires=install_requires,
    license="Apache 2.0"
)

# TODO add classifiers
